"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const fs_1 = require("fs");
const aws_sdk_1 = __importDefault(require("aws-sdk"));
const body_parser_1 = require("body-parser");
const command_line_args_1 = __importDefault(require("command-line-args"));
const cors_1 = __importDefault(require("cors"));
const express_1 = __importDefault(require("express"));
const AlbRequest_1 = require("./model/AlbRequest");
/**
 * Simple console application that hosts an express HTTP
 * server that simulates an AWS ALB.
 *
 * A request is mapped from the HTTP request to a lambda
 * target, the response from invoking the lambda is mapped to
 * a standard HTTP response and returned to the express client.
 */
class AlbApp {
    constructor() {
        this.expressApp = express_1.default();
        this.debugEnabled = false;
    }
    /**
     * Starts the express server.
     *
     * @param args Command line arguments for this server, equivalent to `process.argv.splice(2)`.
     *             See --help for more info.
     */
    async runServer(args) {
        let options = this.parseArguments(args);
        this.debugEnabled = options.debug;
        let listenOnAllHosts = options.host === "*";
        let baseUrl = `http://${options.host}:${options.port}`;
        this.config = this.readConfig(options.config);
        this.configureAws();
        this.configureServer(options);
        this.setupAlbTargetListeners();
        // health check endpoint for testing
        this.expressApp.get("/", (_, res) => res.status(204).send());
        this.expressApp.all("*", (req, res) => {
            let errMessage = `Request does not match any configured ALB target group: ${req.path}`;
            if (this.debugEnabled) {
                this.log(errMessage);
            }
            res.status(400).send({
                error: "Request does not match any configured ALB target group"
            });
        });
        await this.startServer(options.host, options.port, listenOnAllHosts);
        this.log(`Listening for HTTP requests on ${baseUrl} ...`);
    }
    log(message, ...args) {
        console.log(`${new Date().toISOString()} lambda-alb - ${message}`, ...args);
    }
    parseArguments(args) {
        if (args && (args.includes("--debug") || args.includes("-d"))) {
            this.log("Command line arguments: %s", args);
        }
        let options = command_line_args_1.default(AlbApp.APP_OPTIONS, {
            argv: args
        });
        if (!options.config || options.config.trim() === "") {
            throw new Error("--config or -c option must be specified");
        }
        if (!fs_1.existsSync(options.config)) {
            throw new Error(`Config file '${options.config}' not found`);
        }
        return options;
    }
    readConfig(configPath) {
        return JSON.parse(fs_1.readFileSync(configPath, {
            encoding: "UTF8"
        }));
    }
    configureServer(options) {
        this.log("CORS origin set to: %s", options.corsOrigin);
        this.expressApp.use(cors_1.default({
            origin: options.corsOrigin
        }));
        // setup body parser to capture request bodies
        this.expressApp.use(body_parser_1.raw({
            limit: AlbApp.MAX_REQUEST_BODY_SIZE,
            type: r => true
        }));
        this.log("WARNING: To simulate an AWS ALB or Application Gateway the max size for requests is limited to %s", AlbApp.MAX_REQUEST_BODY_SIZE);
    }
    configureAws() {
        let config = {};
        if (this.config.region && this.config.region.trim() !== "") {
            config.region = this.config.region;
        }
        aws_sdk_1.default.config.update(config);
        if (this.config.lambdaEndpoint && this.config.lambdaEndpoint.trim() !== "") {
            config.endpoint = this.config.lambdaEndpoint;
        }
        this.lambdaClient = new aws_sdk_1.default.Lambda(config);
    }
    setupAlbTargetListeners() {
        let self = this;
        for (let targetKey in this.config.targets) {
            if (!this.config.targets.hasOwnProperty(targetKey)) {
                continue;
            }
            let target = this.config.targets[targetKey];
            let basePath = target.routeUrl ? target.routeUrl : `/${targetKey}`;
            if (basePath === "/") {
                throw new Error(`Invalid route URL '/' for lambda target '${targetKey}'`);
            }
            this.log("ALB target configured for lambda '%s' @ route: %s", target.lambdaName, basePath);
            this.expressApp.all(`${basePath}*`, (req, res) => self.handleHttpRequest(self, target, basePath, req, res));
        }
    }
    async handleHttpRequest(self, target, basePath, request, response) {
        try {
            let apiRequestEvent = await self.mapRequestToApiEvent(request, basePath);
            let apiResponse = await self.run(target, apiRequestEvent, {});
            self.forwardApiResponse(apiResponse, response);
        }
        catch (ex) {
            if (this.debugEnabled) {
                this.log("Target lambda '%s' invocation error", target.lambdaName);
                console.dir(ex);
            }
            response.status(503)
                .send({
                error: `Target lambda '${target.lambdaName}' invocation error:\n${JSON.stringify(ex)}`
            });
        }
    }
    async mapRequestToApiEvent(request, basePath) {
        this.log("Mapping express request to AWS model");
        let apiRequest = new AlbRequest_1.AlbRequest();
        apiRequest.httpMethod = request.method;
        apiRequest.path = request.path.substr(basePath.length);
        if (apiRequest.path.trim() === "") {
            apiRequest.path = "/";
        }
        Object.keys(request.headers)
            .forEach(h => apiRequest.headers[h] = request.headers[h]);
        Object.keys(request.query)
            .forEach(q => apiRequest.queryStringParameters[q] = request.query[q]);
        if (!AlbApp.HTTP_METHODS_WITH_ENTITY.includes(request.method)) {
            return apiRequest;
        }
        let body = request.body;
        apiRequest.body = body.toString("base64");
        apiRequest.isBase64Encoded = true;
        return apiRequest;
    }
    async run(target, request, context) {
        let contextBuffer = Buffer.from(JSON.stringify(context));
        let lambaRequest = {
            ClientContext: contextBuffer.toString("base64"),
            FunctionName: target.lambdaName,
            InvocationType: "RequestResponse",
            LogType: "None",
            Payload: JSON.stringify(request),
            Qualifier: target.versionOrAlias
        };
        this.log("Invoking AWS Lambda '%s'%s. Path: %s", target.lambdaName, target.versionOrAlias ? `, using qualifier '${target.versionOrAlias}'` : "", request.path);
        if (this.debugEnabled) {
            let loggableRequest = JSON.parse(JSON.stringify(request));
            if (loggableRequest.isBase64Encoded) {
                loggableRequest.body = `${loggableRequest.body.substr(0, 32)}...`;
            }
            console.dir(loggableRequest);
        }
        return await (new Promise((resolve, reject) => {
            this.lambdaClient.invoke(lambaRequest, (error, data) => {
                if (error) {
                    reject(error);
                    return;
                }
                resolve(JSON.parse(data.Payload));
            });
        }));
    }
    forwardApiResponse(apiResponse, response) {
        this.log("Mapping AWS response model to express response");
        if (this.debugEnabled) {
            let loggableResponse = JSON.parse(JSON.stringify(apiResponse));
            if (loggableResponse.isBase64Encoded) {
                loggableResponse.body = `${loggableResponse.body.substr(0, 32)}...`;
            }
            console.dir(loggableResponse);
        }
        let headers = apiResponse.headers;
        response.status(apiResponse.statusCode);
        Object.keys(headers).forEach(h => response.header(h, headers[h]));
        if (apiResponse.isBase64Encoded) {
            response.contentType(apiResponse.headers["content-type"] || "application/octet-stream");
            response.end(Buffer.from(apiResponse.body, "base64"));
        }
        else {
            response.send(apiResponse.body);
        }
    }
    async startServer(host, port, listenOnAllHosts) {
        await new Promise((resolve, reject) => {
            try {
                if (listenOnAllHosts) {
                    this.log("Listening on all hosts");
                    this.server = this.expressApp.listen(port, resolve);
                }
                else {
                    this.log("Listening on host: %s", host);
                    this.server = this.expressApp.listen(port, host, resolve);
                }
            }
            catch (ex) {
                reject(ex);
            }
        });
    }
    async stopServer() {
        if (!this.server) {
            throw new Error("stopServer can only be called after runServer has been called and has completed");
        }
        this.log("Server shutting down");
        await new Promise((resolve, reject) => {
            try {
                this.server.close(() => resolve());
            }
            catch (ex) {
                reject(ex);
            }
        });
    }
}
AlbApp.MAX_REQUEST_BODY_SIZE = "1024kb";
AlbApp.HTTP_METHODS_WITH_ENTITY = ["POST", "PUT", "PATCH"];
AlbApp.APP_OPTIONS = [{
        alias: "c",
        description: "Path to the JSON configuration file",
        name: "config",
        type: String
    }, {
        alias: "h",
        defaultValue: "*",
        description: "Host to listen on (defaults to all)",
        name: "host",
        type: String
    }, {
        alias: "p",
        defaultValue: 8080,
        description: "Port to listen on (defaults to 8080)",
        name: "port",
        type: Number
    }, {
        alias: "o",
        defaultValue: "*",
        description: "CORS origin to accept (default is to accept any origin)",
        name: "corsOrigin",
        type: String
    }, {
        alias: "d",
        defaultValue: false,
        description: "Enable debug logging",
        name: "debug",
        type: Boolean
    }
];
exports.AlbApp = AlbApp;
//# sourceMappingURL=AlbApp.js.map